use anyhow::Result;
use rosc::{
    OscMessage, OscPacket, OscType,
    decoder::decode_udp,
    encoder::{encode, encode_tcp},
};
use tokio::{
    io::{AsyncReadExt, AsyncWriteExt},
    net::{TcpListener, TcpStream, UdpSocket},
    spawn,
    sync::mpsc::{Receiver, Sender},
    task::JoinSet,
};

use crate::Args;

pub async fn setup_carla(args: Args, tx: Sender<OscPacket>) -> Result<()> {
    let mut set = JoinSet::new();
    set.spawn(tcp_client(args.clone(), tx.clone()));
    set.spawn(udp_client(args, tx));

    while let Some(res) = set.join_next().await {
        res??;
    }

    Ok(())
}

pub async fn setup_forwarder(args: Args, rx: Receiver<OscPacket>) -> Result<()> {
    let mut set = JoinSet::new();
    set.spawn(udp_sender(args, rx));

    while let Some(res) = set.join_next().await {
        res??;
    }

    Ok(())
}

async fn tcp_client(args: Args, tx: Sender<OscPacket>) -> Result<()> {
    println!("Starting TCP carla client");
    let sock = TcpListener::bind(format!("0.0.0.0:{}", args.carla_recv_port))
        .await
        .unwrap();
    register_tcp(args).await?;

    loop {
        let (stream, _) = sock.accept().await?;
        spawn(tcp_connection(stream, tx.clone()));
    }
}

async fn register_tcp(args: Args) -> Result<()> {
    let mut stream = TcpStream::connect(args.carla_host).await?;

    let register = OscMessage {
        addr: String::from("/register"),
        args: vec![OscType::String(format!(
            "osc.tcp://{}:{}/Carla",
            args.bridge_host, args.carla_recv_port
        ))],
    };
    let packet = OscPacket::Message(register);
    let packet = encode_tcp(&packet)?;
    stream.write_all(&packet).await?;

    stream.shutdown().await?;

    Ok(())
}

async fn tcp_connection(stream: TcpStream, tx: Sender<OscPacket>) -> Result<()> {
    println!("Accepted new Carla TCP connection");
    let mut buf = vec![0; 1024];
    let mut stream = stream;

    loop {
        let bytes = stream.read(&mut buf).await?;
        let packet = &buf[0..bytes];
        if bytes > 4 {
            let (_, packet) = decode_udp(packet)?;
            tx.send(packet).await?;
        }
    }
}

async fn udp_client(args: Args, tx: Sender<OscPacket>) -> Result<()> {
    println!("Starting UDP carla client");
    let mut sock = UdpSocket::bind(format!("0.0.0.0:{}", args.carla_recv_port))
        .await
        .unwrap();
    register_udp(args, &mut sock).await?;

    let mut buf = vec![0; 1024];

    loop {
        let bytes = sock.recv(&mut buf).await?;
        let packet = &buf[0..bytes];
        let (_, packet) = decode_udp(packet)?;
        tx.send(packet).await?;
    }
}

async fn register_udp(args: Args, sock: &mut UdpSocket) -> Result<()> {
    let register = OscMessage {
        addr: String::from("/register"),
        args: vec![OscType::String(format!(
            "osc.tcp://{}:{}/Carla",
            args.bridge_host, args.carla_recv_port
        ))],
    };
    let packet = OscPacket::Message(register);
    let packet = encode(&packet)?;
    sock.connect(args.carla_host).await?;
    sock.send(&packet).await?;

    Ok(())
}

async fn udp_sender(args: Args, rx: Receiver<OscPacket>) -> Result<()> {
    let sock = UdpSocket::bind(format!("0.0.0.0:{}", args.carla_send_port))
        .await
        .unwrap();
    sock.connect(args.carla_host).await?;
    let mut rx = rx;

    loop {
        let packet = rx.recv().await;
        match packet {
            Some(packet) => {
                let packet = encode(&packet)?;
                sock.send(&packet).await?;
            }
            None => {
                println!("No more packets");
                return Ok(());
            }
        };
    }
}
