use anyhow::Result;
use rosc::{OscBundle, OscMessage, OscPacket, decoder::decode_udp};
use tokio::{net::UdpSocket, sync::mpsc::Sender, task::JoinSet};

use crate::Args;

pub async fn setup_server(args: Args, tx: Sender<OscPacket>) -> Result<()> {
    println!("Starting server");

    let mut set = JoinSet::new();
    set.spawn(server(args, tx));

    while let Some(res) = set.join_next().await {
        res??;
    }

    Ok(())
}

async fn server(args: Args, tx: Sender<OscPacket>) -> Result<()> {
    println!("Starting UDP OSC Server");
    let sock = UdpSocket::bind(format!("0.0.0.0:{}", args.client_recv_port))
        .await
        .unwrap();

    let mut buf = vec![0; 1024];

    loop {
        let bytes = sock.recv(&mut buf).await?;
        let packet = &buf[0..bytes];
        let (_, packet) = decode_udp(packet)?;
        if let Some(packet) = translate_packet(packet) {
            tx.send(packet).await?;
        }
    }
}

fn translate_packet(packet: OscPacket) -> Option<OscPacket> {
    match packet {
        OscPacket::Message(ref msg) => match msg.addr.as_str() {
            "/Carla/param" => Some(OscPacket::Message(OscMessage {
                addr: format!("/Carla/{}/set_parameter_value", msg.args[0].clone().int()?),
                args: vec![msg.args[1].clone(), msg.args[2].clone()],
            })),
            _ => None,
        },
        OscPacket::Bundle(OscBundle { timetag, content }) => {
            let content = content.into_iter().filter_map(translate_packet).collect();
            Some(OscPacket::Bundle(OscBundle { timetag, content }))
        }
    }
}
